from ms_deisotope.peak_set import DeconvolutedPeakSet
from ms_deisotope.spectrum_graph import (PathFinder, amino_acids, MassWrapper)

from brainpy import SimpleComposition


H2O = SimpleComposition({"H": 2, "O": 1})
NH3 = SimpleComposition({"H": 3, "N": 1})

default_losses = [MassWrapper("-H2O", -H2O.mass()), MassWrapper("-NH3", -NH3.mass())]


class GraphBasedDenoiser(object):
    def __init__(self, components=None, precursor_error_tolerance=2e-5, product_error_tolerance=2e-5,
                 precursor_losses=None, product_losses=None):
        if components is None:
            components = amino_acids
        if precursor_losses is None:
            precursor_losses = default_losses[:]
        if product_losses is None:
            product_losses = default_losses[:]
        self.precursor_error_tolerance = precursor_error_tolerance
        self.product_error_tolerance = product_error_tolerance
        self.path_finder = PathFinder(components, product_error_tolerance)
        self.components = self.path_finder.components
        self.precursor_losses = list(precursor_losses)
        self.product_losses = list(product_losses)

    def initialize_graph(self, peak_set, precursor_mass):
        """Construct a spectrum graph which connects peaks by :attr:`component` masses

        This method relies on :attr:`path_finder` for most of its heavy lifting through
        :meth:`~.PathFinder.find_edges` and :meth:`~.PathFinder.find_complements`.

        Parameters
        ----------
        peak_set : :class:`~.DeconvolutedPeakSet`
            The peak set to search
        precursor_mass : float
            The intact mass of the precursor ion

        Returns
        -------
        :class:`~.SpectrumGraph`
        """
        graph = self.path_finder.find_edges(peak_set)
        p = MassWrapper("Precursor", precursor_mass)
        self.path_finder.find_complements(peak_set, graph, p)
        return graph

    def find_precursor_peak(self, peak_set, precursor_mass, losses=None):
        """Find peaks derived from intact- and neutral losses of the precursor ion

        Parameters
        ----------
        peak_set : :class:`~.DeconvolutedPeakSet`
            The peak set to search
        precursor_mass : float
            The intact mass of the precursor ion
        losses : :class:`~.Iterable` of :class:`~.MassWrapper`, optional
            An iterable of losses expressed as :class:`~.MassWrapper` instances. If not
            provided, -H2O and -NH3 will be assumed.

        Returns
        -------
        :class:`set`:
            The indices of the peaks derived from the intact precursor to retain
        """
        if losses is None:
            losses = self.precursor_losses
        peaks = peak_set.all_peaks_for(precursor_mass, self.product_error_tolerance)
        keep = set()
        for peak in peaks:
            keep.add(peak.index.neutral_mass)
        for loss in losses:
            peaks = peak_set.all_peaks_for(
                precursor_mass + loss.mass, self.product_error_tolerance)
            for peak in peaks:
                keep.add(peak.index.neutral_mass)
        return keep

    def find_loss_peaks(self, peak_set, kept_peak_indices, losses=None):
        """Find neutral loss peaks from the set of peaks that are already set to be retained.

        Parameters
        ----------
        peak_set : :class:`~.DeconvolutedPeakSet`
            The peak set to search
        kept_peak_indices : :class:`set`
            The indices of the peaks already retained to find losses for.
        losses : :class:`~.Iterable` of :class:`~.MassWrapper`, optional
            An iterable of losses expressed as :class:`~.MassWrapper` instances. If not
            provided, -H2O and -NH3 will be assumed.

        Returns
        -------
        :class:`set`:
            `kept_peak_indices` is updated and returned
        """
        if losses is None:
            losses = self.product_losses
        for index in tuple(kept_peak_indices):
            peak = peak_set[index]
            for loss in losses:
                for loss_peak in peak_set.all_peaks_for(
                        peak.neutral_mass + loss.mass, self.product_error_tolerance):
                    kept_peak_indices.add(loss_peak.index.neutral_mass)
        return kept_peak_indices

    def walk_graph(self, graph):
        """Traverse `graph` along its edges, and save the peak indices at either end of each
        edge for use later.

        [description]

        Parameters
        ----------
        graph : :class:`~.SpectrumGraph`
            The graph built by :meth:`initialize_graph`

        Returns
        -------
        :class:`set`:
            The set of peaks which were joined by edges.
        """
        keep = set()
        for e in graph.transitions:
            keep.add(e.start.index)
            keep.add(e.end.index)
        return keep

    def select_peaks(self, peak_set, indices):
        """Given a :class:`~.DeconvolutedPeakSet` and a set of indices into it, construct a
        new :class:`~.DeconvolutedPeakSet` with copies of the peaks at those indices

        Parameters
        ----------
        peak_set : :class:`~.DeconvolutedPeakSet`
            The peak set to search
        indices : :class:`set`
            The indices of the peaks to retain

        Returns
        -------
        :class:`~.DeconvolutedPeakSet`
        """
        keep = []
        for i in indices:
            keep.append(peak_set[i].clone())
        keep = DeconvolutedPeakSet(keep)
        keep.reindex()
        return keep

    def apply_filter(self, peak_set, precursor_mass):
        """Does the actual work of filtering the peak set.

        This method is designed to decouple the :class:`~.ScanBase`
        consuming interface from the actual filtering process.

        Parameters
        ----------
        peak_set : :class:`~.DeconvolutedPeakSet`
            The peak set to search
        precursor_mass : float
            The intact mass of the precursor ion

        Returns
        -------
        :class:`~.DeconvolutedPeakSet`

        See Also
        --------
        :meth:`filter`
        :meth:`__call__`
        """
        graph = self.initialize_graph(peak_set, precursor_mass)
        indices = self.walk_graph(graph)
        self.find_loss_peaks(peak_set, indices)
        indices.update(self.find_precursor_peak(peak_set, precursor_mass))
        peak_set = self.select_peaks(peak_set, indices)
        return peak_set

    def filter(self, scan):
        """Filter an MSn scan's deconvoluted peak set according to graph-connectedness.

        This noise filter will remove any peak that is not connected to another peak by a
        mass shift from :attr:`components` or connected to such a peak via a loss mass shift
        given by :attr:`product_losses` or :attr:`precursor_losses`

        This method is also used via the :meth:`__call__` slot.

        Parameters
        ----------
        scan : :class:`~.ScanBase`
            The MSn scan to filter.

        Returns
        -------
        :class:`~.DeconvolutedPeakSet`

        See Also
        --------
        :meth:`apply_filter`
        """
        precursor_mass = scan.precursor_information.extracted_neutral_mass
        if precursor_mass == 0:
            precursor_mass = scan.precursor_information.neutral_mass
        peak_set = scan.deconvoluted_peak_set
        if peak_set is None:
            raise ValueError("scan.deconvoluted_peak_set cannot be None!")
        filtered_peak_set = self.apply_filter(peak_set, precursor_mass)
        return filtered_peak_set

    def __call__(self, scan):
        """Filter an MSn scan's deconvoluted peak set according to graph-connectedness.

        This is a wrapper around the :meth:`filter` method.

        Parameters
        ----------
        scan : :class:`~.ScanBase`
            The MSn scan to filter.

        Returns
        -------
        :class:`~.DeconvolutedPeakSet`

        See Also
        --------
        :meth:`filter`
        """
        return self.filter(scan)



# cython: embedsignature=True

cimport cython
from libc.math cimport fabs, sqrt, log, ceil, floor
from libc.stdlib cimport malloc, free
import operator

from cpython.list cimport PyList_New, PyList_GetItem, PyList_Size, PyList_GET_ITEM, PyList_SET_ITEM, PyList_GET_SIZE
from cpython.sequence cimport PySequence_List

from brainpy._c.isotopic_distribution cimport TheoreticalPeak
from ms_peak_picker._c.peak_set cimport PeakSet, FittedPeak


@cython.nonecheck(False)
@cython.cdivision(True)
cdef bint isclose(double x, double y, double rtol=1.e-5, double atol=1.e-8):
    return abs(x-y) <= (atol + rtol * abs(y))


@cython.freelist(50000)
cdef class IsotopicFitRecord(object):
    """Describes a single isotopic pattern fit, comparing how well an
    experimentally observed sequence of peaks matches a theoretical isotopic
    pattern.

    IsotopicFitRecord instances are hashable and orderable (by score).

    Attributes
    ----------
    charge : int
        The charge state used to generate the theoretical pattern
    data : object
        An arbitrary Python object containing extra information
    experimental : list of FittedPeak
        The observed experimental peaks to be fitted
    missed_peaks : int
        The number of peaks in the theoretical pattern that do not
        have a matching experimental peak
    monoisotopic_peak : FittedPeak
        The fitted peak which corresponds to the monoisotopic peak
    score : float
        The score assigned to the fit by an IsotopicFitter object
    seed_peak : FittedPeak
        The peak that was used to initiate the fit. This may be unused
        if not using an Averagine method
    theoretical : TheoreticalIsotopicPattern
        The theoretical isotopic pattern being fitted on the experimental data
    """
    def __init__(self, FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical, list experimental,
                  object data=None, int missed_peaks=0):
        self.seed_peak = seed_peak
        self.score = score
        self.charge = charge
        self.experimental = experimental
        self.theoretical = theoretical
        self.monoisotopic_peak = experimental[0]
        self.data = data
        self.missed_peaks = missed_peaks
        self._hash = -1

    @staticmethod
    cdef IsotopicFitRecord _create(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical,
                                   list experimental, object data=None, int missed_peaks=0):
        cdef IsotopicFitRecord inst
        inst = IsotopicFitRecord.__new__(IsotopicFitRecord)
        inst.seed_peak = seed_peak
        inst.score = score
        inst.charge = charge
        inst.theoretical = theoretical
        inst.experimental = experimental
        inst.monoisotopic_peak = <FittedPeak>PyList_GetItem(experimental, 0)
        inst.data = data
        inst.missed_peaks = missed_peaks
        inst._hash = -1
        return inst

    def clone(self):
        return IsotopicFitRecord._create(self.seed_peak, self.score, self.charge, self.theoretical,
                              self.experimental, self.data, self.missed_peaks)

    def __reduce__(self):
        return self.__class__, (self.seed_peak, self.score, self.charge, self.theoretical,
                                self.experimental, self.data, self.missed_peaks)

    cpdef bint _eq(self, IsotopicFitRecord other):
        cdef:
            bint val
            size_t i, n
            FittedPeak fp1, fp2

        val = self.score == other.score
        if not val:
            return val
        val = self.charge == other.charge
        if not val:
            return val
        n = PyList_GET_SIZE(self.experimental)
        val = n == PyList_GET_SIZE(other.experimental)
        if not val:
            return val
        for i in range(n):
            fp1 = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
            fp2 = <FittedPeak>PyList_GET_ITEM(other.experimental, i)
            val = fp1._eq(fp2)
            if not val:
                return val
        val = self.theoretical._eq_inst(other.theoretical)
        if not val:
            return val
        if self.data is not None or other.data is not None:
            val = (self.data == other.data)
        return val

    cpdef bint _ne(self, IsotopicFitRecord other):
        return not (self._eq(other))

    cpdef bint _lt(self, IsotopicFitRecord other):
        return self.score < other.score

    cpdef bint _gt(self, IsotopicFitRecord other):
        return self.score > other.score

    def __richcmp__(self, IsotopicFitRecord other, int code):
        if other is None:
            if code == 3:
                return True
            else:
                return False

        if code == 0:
            return self._lt(other)
        elif code == 2:
            return self._eq(other)
        elif code == 3:
            return not (self._eq(other))
        elif code == 4:
            return self._gt(other)


    def __hash__(self):
        if self._hash == -1:
            self._hash = hash((self.monoisotopic_peak.mz, self.charge))
        return self._hash

    def __iter__(self):
        yield self.score
        yield self.charge
        yield self.experimental
        yield self.theoretical

    @property
    def npeaks(self):
        return len(self.experimental)

    def __repr__(self):
        return "IsotopicFitRecord(score=%0.5f, charge=%d, npeaks=%d, monoisotopic_mz=%0.5f)" % (
            self.score, self.charge, self.npeaks, self.monoisotopic_peak.mz)


cdef class FitSelectorBase(object):
    """An object that controls the filtering and
    selection of IsotopicFitRecord

    Attributes
    ----------
    minimum_score : int
        The minimum score needed to be a candidate for selection. If the
        FitSelector is `minimizing` it is the maximal score to be a candidate.
    """
    def __init__(self, minimum_score=0):
        self.minimum_score = minimum_score

    def __reduce__(self):
        return self.__class__, (self.minimum_score,)

    cpdef IsotopicFitRecord best(self, object results):
        raise NotImplementedError()

    cdef IsotopicFitRecord _best_from_set(self, set results):
        raise NotImplementedError()

    def __call__(self, *args, **kwargs):
        return self.best(*args, **kwargs)

    cpdef bint reject(self, IsotopicFitRecord result):
        raise NotImplementedError()

    cpdef bint reject_score(self, double score):
        raise NotImplementedError()

    cpdef bint is_maximizing(self):
        return False

    def __repr__(self):
        return "{self.__class__.__name__}(minimum_score={self.minimum_score})".format(self=self)


cdef class MinimizeFitSelector(FitSelectorBase):
    """A FitSelector which tries to minimize the score of the best fit.
    """
    cpdef IsotopicFitRecord best(self, object results):
        """Returns the IsotopicFitRecord with the smallest score

        Parameters
        ----------
        results : list of IsotopicFitRecord
            List of isotopic fits to select the most optimal case from

        Returns
        -------
        IsotopicFitRecord
            The most optimal fit
        """
        if isinstance(results, set):
            return self._best_from_set(<set>results)
        return min(results, key=operator.attrgetter("score"))

    cdef IsotopicFitRecord _best_from_set(self, set results):
        cdef:
            IsotopicFitRecord fit, best_fit
            object obj
            double best_score

        best_fit = None
        best_score = INFINITY
        for obj in results:
            fit = <IsotopicFitRecord>obj
            if fit.score < best_score:
                best_score = fit.score
                best_fit = fit
        return best_fit


    cpdef bint reject(self, IsotopicFitRecord fit):
        """Decide whether the fit should be discarded for having too
        large a score. Compares against :attr:`minimum_score`

        Parameters
        ----------
        fit : IsotopicFitRecord

        Returns
        -------
        bool
        """
        return fit.score > self.minimum_score

    cpdef bint reject_score(self, double score):
        return score > self.minimum_score

    cpdef bint is_maximizing(self):
        """Returns that this is *not* a maximizing selector

        Returns
        -------
        False
        """
        return False


cdef class MaximizeFitSelector(FitSelectorBase):
    """A FitSelector which tries to maximize the score of the best fit.
    """
    cpdef IsotopicFitRecord best(self, object results):
        """Returns the IsotopicFitRecord with the largest score

        Parameters
        ----------
        results : list of IsotopicFitRecord
            List of isotopic fits to select the most optimal case from

        Returns
        -------
        IsotopicFitRecord
            The most optimal fit
        """
        if isinstance(results, set):
            return self._best_from_set(<set>results)
        return max(results, key=operator.attrgetter("score"))

    cdef IsotopicFitRecord _best_from_set(self, set results):
        cdef:
            IsotopicFitRecord fit, best_fit
            double best_score

        best_fit = None
        best_score = -INFINITY
        for obj in results:
            fit = <IsotopicFitRecord>obj
            if fit.score > best_score:
                best_score = fit.score
                best_fit = fit
        return best_fit

    cpdef bint reject(self, IsotopicFitRecord fit):
        """Decide whether the fit should be discarded for having too
        small a score. Compares against :attr:`minimum_score`

        Parameters
        ----------
        fit : IsotopicFitRecord

        Returns
        -------
        bool
        """
        return fit.score < self.minimum_score

    cpdef bint reject_score(self, double score):
        return score < self.minimum_score

    cpdef bint is_maximizing(self):
        """Returns that this *is* a maximizing selector

        Returns
        -------
        False
        """
        return True


cdef class IsotopicFitterBase(object):
    """A base class for Isotopic Pattern Fitters, objects
    which given a set of experimental peaks and a set of matching
    theoretical peaks, returns a fit score describing how good the
    match is.

    An IsotopicFitter may be optimal when the score is small (minimizing)
    or when the score is large (maximizing), and the appropriate
    :class:`FitSelectorBase` type will be used for :attr:`select`. This
    will also be reflected by :meth:`is_maximizing`.
    """

    def __init__(self, score_threshold=0.5):
        self.select = MinimizeFitSelector(score_threshold)

    def __reduce__(self):
        return self.__class__, (0,), self.__getstate__()

    def __getstate__(self):
        return (self.select,)

    def __setstate__(self, state):
        self.select, = state

    def evaluate(self, PeakSet peaklist, list observed, list expected, **kwargs):
        """Evaluate a pair of peak lists for goodness-of-fit.

        Parameters
        ----------
        peaklist : :class:`~.PeakSet`
            The full set of all experimental peaks
        observed : list
            The list of experimental peaks that are part of this fit
        expected : list
            The list of theoretical peaks that are part of this fit
        **kwargs

        Returns
        -------
        float
            The score
        """
        return self._evaluate(peaklist, observed, expected)

    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        raise NotImplementedError()

    def __call__(self, *args, **kwargs):
        """Invokes :meth:`evaluate`

        Parameters
        ----------
        *args
            Forwarded to :meth:`evaluate`
        **kwargs
            Forwarded to :meth:`evaluate`

        Returns
        -------
        float
            The score
        """
        return self.evaluate(*args, **kwargs)

    cpdef bint reject(self, IsotopicFitRecord fit):
        """Test whether this fit is too poor to be used

        Parameters
        ----------
        fit : :class:`~IsotopicFitRecord`
            The fit to test

        Returns
        -------
        bool
        """
        return self.select.reject(fit)

    cpdef bint reject_score(self, double score):
        return self.select.reject_score(score)

    cpdef bint is_maximizing(self):
        """Whether or not this fitter's score gets better as it grows

        Returns
        -------
        bool
            Whether or not this fitter is a maximizing fitter
        """
        return self.select.is_maximizing()

    cpdef IsotopicFitterBase _configure(self, DeconvoluterBase deconvoluter, dict kwargs):
        return self

    def configure(self, DeconvoluterBase deconvoluter, **kwargs):
        return self._configure(deconvoluter, kwargs)

    def __repr__(self):
        return "{self.__class__.__name__}({fields})".format(self=self, fields=self.__getstate__())


cdef double sum_intensity_theoretical(list peaklist):
    cdef:
        double summed
        size_t i
        TheoreticalPeak peak
    summed = 0
    for i in range(PyList_GET_SIZE(peaklist)):
        peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
        summed += peak.intensity
    return summed


cdef double sum_intensity_fitted(list peaklist):
    cdef:
        double summed
        size_t i
        FittedPeak peak
    summed = 0
    for i in range(PyList_GET_SIZE(peaklist)):
        peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
        summed += peak.intensity
    return summed


@cython.cdivision
cdef double* normalize_intensity_theoretical(list peaklist, double* out, double total):
    cdef:
        size_t i
        TheoreticalPeak peak

    for i in range(PyList_GET_SIZE(peaklist)):
        peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
        out[i] = peak.intensity / total


@cython.cdivision
cdef double* normalize_intensity_fitted(list peaklist, double* out, double total):
    cdef:
        size_t i
        FittedPeak peak

    for i in range(PyList_GET_SIZE(peaklist)):
        peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
        out[i] = peak.intensity / total


cdef class GTestFitter(IsotopicFitterBase):
    r"""Evaluate an isotopic fit using a G-test

    .. math::
        G = 2\sum_i^n{o_i * ({log}o_i - {log}e_i)}

    where :math:`o_i` is the intensity of the ith experimental peak
    and :math:`e_i` is the intensity of the ith theoretical peak.

    .. note::

        The G statistic is on the scale of the signal used.
    """
    @cython.cdivision
    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        cdef:
            double g_score, obs, theo, log_ratio
            size_t n

        n = PyList_GET_SIZE(observed)

        g_score = 0.
        for i in range(n):
            obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity
            theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity

            log_ratio = log(obs) - log(theo)
            g_score += obs * log_ratio

        return g_score * 2.


cdef class ScaledGTestFitter(IsotopicFitterBase):
    r"""Evaluate an isotopic fit using a G-test after normalizing the
    list of experimental and theoretical peaks to both sum to 1.

    .. math::
        G = 2\sum_i^n{o_i * ({log}o_i - {log}e_i)}

    where :math:`o_i` is the intensity of the ith experimental peak
    and :math:`e_i` is the intensity of the ith theoretical peak.
    """

    @cython.cdivision
    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        cdef:
            double total_observed
            double total_expected
            double g_score, obs, theo, log_ratio
            size_t n
        total_observed = sum_intensity_fitted(observed)
        total_expected = sum_intensity_theoretical(expected)
        n = PyList_GET_SIZE(observed)

        g_score = 0.
        for i in range(n):
            obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity / total_observed
            theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity / total_expected

            log_ratio = log(obs) - log(theo)
            g_score += obs * log_ratio

        return g_score * 2.


cdef ScaledGTestFitter g_test_scaled

g_test_scaled = ScaledGTestFitter()


cdef double max_intensity_theoretical(list peaklist):
    cdef:
        double maximum
        size_t i
        TheoreticalPeak peak
    maximum = 0
    for i in range(PyList_GET_SIZE(peaklist)):
        peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
        if peak.intensity > maximum:
            maximum = peak.intensity
    return maximum


cdef double max_intensity_fitted(list peaklist):
    cdef:
        double maximum
        size_t i
        FittedPeak peak
    maximum = 0
    for i in range(PyList_GET_SIZE(peaklist)):
        peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
        if peak.intensity > maximum:
            maximum = peak.intensity
    return maximum


cdef class LeastSquaresFitter(IsotopicFitterBase):
    r"""Evaluate an isotopic fit using least squares coefficient of
    determination :math:`R^2`.

    .. math::
        {\hat e_i} &= e_i / max(e)

        {\hat t_i} &= t_i / max(t)

        {\hat t} &= \sum_i^n {\hat t_i}^2

        R^2 &= \frac{1}{{\hat t}}\sum_i^n ({\hat e_i} - {\hat t_i})^2

    where :math:`e_i` is the ith experimental peak intensity and :math:`t_i`
    is the ith theoretical peak intensity
    """
    @cython.cdivision
    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        cdef:
            double exp_max, theo_max, sum_of_squared_errors, sum_of_squared_theoreticals
            double normed_theo, normed_expr
            size_t i
            TheoreticalPeak t
            FittedPeak e

        exp_max = max_intensity_fitted(observed)
        theo_max = max_intensity_theoretical(expected)

        sum_of_squared_errors = 0
        sum_of_squared_theoreticals = 0

        assert len(observed) == len(expected)

        # for e, t in zip(observed, expected):
        for i in range(PyList_GET_SIZE(observed)):
            e = <FittedPeak>PyList_GetItem(observed, i)
            t = <TheoreticalPeak>PyList_GetItem(expected, i)
            normed_expr = e.intensity / exp_max
            normed_theo = t.intensity / theo_max
            sum_of_squared_errors += (normed_theo - normed_expr) ** 2
            sum_of_squared_theoreticals += normed_theo ** 2
        return sum_of_squared_errors / sum_of_squared_theoreticals


cdef LeastSquaresFitter least_squares

least_squares = LeastSquaresFitter()


cdef class ChiSquareFitter(IsotopicFitterBase):

    @cython.cdivision
    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        cdef:
            double running_total
            size_t i
            TheoreticalPeak t
            FittedPeak e
        running_total = 0
        for i in range(PyList_GET_SIZE(observed)):
            e = <FittedPeak>PyList_GetItem(observed, i)
            t = <TheoreticalPeak>PyList_GetItem(expected, i)
            running_total += ((e - t) ** 2) / t
        return running_total


@cython.cdivision
cdef double ms_deconv_score_peak(FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=0.02, double minimum_signal_to_noise=1) nogil:
    cdef:
        double mass_error, abundance_diff
    if obs.signal_to_noise < minimum_signal_to_noise:
        return 0.

    mass_error = fabs(obs.mz - theo.mz)

    if mass_error <= mass_error_tolerance:
        mass_accuracy = 1 - mass_error / mass_error_tolerance
    else:
        return 0.

    if obs.intensity < theo.intensity and (((theo.intensity - obs.intensity) / obs.intensity) <= 1):
        abundance_diff = 1 - ((theo.intensity - obs.intensity) / obs.intensity)
    elif obs.intensity >= theo.intensity and (((obs.intensity - theo.intensity) / obs.intensity) <= 1):
        abundance_diff = sqrt(1 - ((obs.intensity - theo.intensity) / obs.intensity))
    else:
        return 0.
    score = sqrt(theo.intensity) * mass_accuracy * abundance_diff
    return score


cdef class MSDeconVFitter(IsotopicFitterBase):
    r"""An implementation of the scoring function used in :title-reference:`MSDeconV`

    .. math::
        :nowrap:

        \begin{split}
            s_{mz}(e, t) &= \begin{cases}
                1 - \frac{\|mz(e) - mz(t)\|}{d} & \text{if } \|mz(e) - mz(t)\| < d,\\
                0 & \text{otherwise}
            \end{cases}\\

            s_{int}(e, t) &= \begin{cases}
                1 - \frac{int(t) - int(e)}{int(e)} & \text{if } int(e) < int(t) \text{ and } \frac{int(t) - int(e)}{int(e)} \le 1, \\
                \sqrt{1 - \frac{int(e) - int(t)}{int(t)}} & \text{if } int(e) \ge int(t) \text{ and } \frac{int(e) - int(t)}{int(t)} \le 1,\\
                0 & \text{otherwise}
            \end{cases}\\

            \text{S}(e, t) &= \sqrt{int(t)}\times s_{mz}(e,t) \times s_{int}(e, t)
        \end{split}

    References
    ----------
    Liu, X., Inbar, Y., Dorrestein, P. C., Wynne, C., Edwards, N., Souda, P., …
    Pevzner, P. A. (2010). Deconvolution and database search of complex tandem
    mass spectra of intact proteins: a combinatorial approach. Molecular & Cellular
    Proteomics : MCP, 9(12), 2772–2782. https://doi.org/10.1074/mcp.M110.002766
    """
    def __init__(self, minimum_score=10, mass_error_tolerance=0.02):
        self.mass_error_tolerance = mass_error_tolerance
        self.select = MaximizeFitSelector()
        self.select.minimum_score = minimum_score

    @cython.cdivision
    cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=0.02, double minimum_signal_to_noise=1) nogil:
        return ms_deconv_score_peak(obs, theo, mass_error_tolerance, minimum_signal_to_noise)

    def evaluate(self, PeakSet peaklist, list observed, list expected):
        return self._evaluate(peaklist, observed, expected)

    def __reduce__(self):
        return self.__class__, (0,), self.__getstate__()

    def __getstate__(self):
        return (self.select, self.mass_error_tolerance)

    def __setstate__(self, state):
        self.select, self.mass_error_tolerance = state

    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        cdef:
            size_t i, n
            FittedPeak obs
            TheoreticalPeak theo
            double score

        n = PyList_GET_SIZE(observed)
        score = 0
        for i in range(n):
            obs = <FittedPeak>PyList_GET_ITEM(observed, i)
            theo = <TheoreticalPeak>PyList_GET_ITEM(expected, i)
            score += self.score_peak(obs, theo, self.mass_error_tolerance, 1)

        return score


cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):
    r"""An Isotopic Fitter which uses the :class:`MSDeconVFitter` score
    weighted by 1 - :attr:`penalty_factor` * :class:`ScaledGTestFitter` score

    .. math::
        S(e, t) = M(e, t) * (1 - G(e, t))

    where :math:`e` is the experimental peak list and :math:`t` is the theoretical
    peak list
    """
    def __init__(self, minimum_score=10, penalty_factor=1, mass_error_tolerance=0.02):
        self.select = MaximizeFitSelector(minimum_score)
        self.penalty_factor = penalty_factor
        self.mass_error_tolerance = mass_error_tolerance

    def __reduce__(self):
        return self.__class__, (0,), self.__getstate__()

    def __getstate__(self):
        return (self.select, self.penalty_factor, self.mass_error_tolerance)

    def __setstate__(self, state):
        self.select, self.penalty_factor, self.mass_error_tolerance = state

    def evaluate(self, PeakSet peaklist, list observed, list expected):
        return self._evaluate(peaklist, observed, expected)

    @cython.cdivision
    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        cdef:
            size_t i, n
            FittedPeak obs
            TheoreticalPeak theo
            double score, total_intensity_observed, total_intensity_expected
            double penalty, _obs, _theo, log_ratio

        total_intensity_observed = 0
        total_intensity_expected = 0

        n = PyList_GET_SIZE(observed)
        score = 0
        for i in range(n):
            obs = <FittedPeak>PyList_GET_ITEM(observed, i)
            theo = <TheoreticalPeak>PyList_GET_ITEM(expected, i)
            score += ms_deconv_score_peak(obs, theo, self.mass_error_tolerance, 1)
            total_intensity_observed += obs.intensity
            total_intensity_expected += theo.intensity
        penalty = 0
        for i in range(n):
            _obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity / total_intensity_observed
            _theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity / total_intensity_expected

            log_ratio = log(_obs) - log(_theo)
            penalty += _obs * log_ratio
        penalty = abs(2 * penalty)
        score *= ((1 - penalty * self.penalty_factor))
        return score


cdef class FunctionScorer(IsotopicFitterBase):
    """Use a user-provided Python function or callable object to evaluate an isotopic
    envelope fit.

    The provided function should take two arguments, a list of experimental fitted peaks
    and a list of theoretical peaks, and return a single number as an output.

    Make sure to pass the right thresholds and selector types when creating an instance
    of this class that match the semantics of the provided function.

    While this type is available for convenience, there is considerable overhead in using
    it compared to one of the C accelerated fitters.
    """

    def __init__(self, function, minimum_score=10, selector_type=MaximizeFitSelector):
        self.function = function
        self.select = selector_type(minimum_score)

    def __getstate__(self):
        return self.select, self.function

    def __setstate__(self, state):
        self.select, self.function = state

    def __reduce__(self):
        return self.__class__, (None,), self.__getstate__()

    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        return self.function(observed, expected)


cdef class InterferenceDetection(object):
    def __init__(self, PeakSet peaklist):
        self.peaklist = peaklist

    def __reduce__(self):
        return self.__class__, (None,), self.__getstate__()

    def __getstate__(self):
        return (self.peaklist,)

    def __setstate__(self, state):
        self.peaklist, = state

    cdef double detect_interference(self, list experimental_peaks, double lower=-1, double upper=-1):
        cdef:
            FittedPeak min_peak, max_peak
            PeakSet region
            double included_intensity, region_intensity, score
            size_t n

        n = PyList_GET_SIZE(experimental_peaks)
        if n == 0:
            return 0.0

        min_peak = <FittedPeak>PyList_GET_ITEM(experimental_peaks, 0)
        max_peak = <FittedPeak>PyList_GET_ITEM(experimental_peaks, PyList_GET_SIZE(experimental_peaks) - 1)

        if lower < 0:
            lower = min_peak.mz - min_peak.full_width_at_half_max
        if upper < 0:
            upper = max_peak.mz + max_peak.full_width_at_half_max

        region = self.peaklist._between(
            lower,
            upper)

        included_intensity = sum_intensity_fitted(experimental_peaks)
        region_intensity = sum_intensity_fitted(list(region))
        if region_intensity == 0:
            return 0.0

        score = 1 - (included_intensity / region_intensity)
        return score

    def __call__(self, list experimental_peaks, lower=None, upper=None):
        if lower is None:
            lower = -1
        if upper is None:
            upper = -1
        return self.detect_interference(experimental_peaks, lower, upper)

cdef class DistinctPatternFitter(IsotopicFitterBase):

    def __init__(self, minimum_score=0.3, peak_count_scale=1.5, domain_scale=100.):
        self.select = MinimizeFitSelector(minimum_score)
        self.interference_detector = None
        self.g_test_scaled = ScaledGTestFitter()
        self.peak_count_scale = peak_count_scale
        self.domain_scale = domain_scale

    def __getstate__(self):
        return self.select, self.interference_detector, self.g_test_scaled, self.peak_count_scale, self.domain_scale

    def __setstate__(self, state):
        self.select, self.interference_detector, self.g_test_scaled, self.peak_count_scale, self.domain_scale = state

    cpdef IsotopicFitterBase _configure(self, DeconvoluterBase deconvoluter, dict kwargs):
        self.interference_detector = InterferenceDetection(deconvoluter.peaklist)
        return self

    cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):
        cdef:
            double score
            double npeaks

        npeaks = PyList_GET_SIZE(experimental)

        if self.interference_detector is None:
            self.interference_detector = InterferenceDetection(peaklist)

        score = self.g_test_scaled._evaluate(peaklist, experimental, theoretical)
        score *= abs((self.interference_detector.detect_interference(experimental) + 0.00001) / (
            npeaks * self.peak_count_scale)) * self.domain_scale
        return score


cdef double percentile(double[:] N, double percent):
    cdef:
        double k, f, c, d0, d1
    k = (N.shape[0] - 1) * percent
    f = floor(k)
    c = ceil(k)
    if f == c:
        return N[<size_t>(k)]
    d0 = N[<size_t>(f)] * (c - k)
    d1 = N[<size_t>(c)] * (k - f)
    return d0 + d1


cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):

    def __init__(self,  minimum_score=0.3, penalty_factor=1., mass_error_tolerance=0.02):
        self.select = MaximizeFitSelector(minimum_score)
        self.scale_factor = 0.
        self.scorer = PenalizedMSDeconVFitter(
            penalty_factor=penalty_factor, mass_error_tolerance=mass_error_tolerance)

    def __getstate__(self):
        return self.select, self.scale_factor, self.scorer

    def __setstate__(self, state):
        self.select, self.scale_factor, self.scorer = state

    @property
    def penalty_factor(self):
        return self.scorer.penalty_factor

    @penalty_factor.setter
    def penalty_factor(self, value):
        self.scorer.penalty_factor = value

    cpdef double _calculate_scale_factor(self, PeakSet peaklist):
        cdef:
            size_t i
            FittedPeak peak
            double maximum
            double intensity
            PeakSet peaks

        maximum = 0.
        peaks = peaklist.peaks
        for i in range(len(peaks)):
            peak = peaks.getitem(i)
            if peak.intensity > maximum:
                maximum = peak.intensity
        return maximum

    cdef void scale_fitted_peaks(self, list experimental, double factor):
        cdef:
            size_t i
            FittedPeak peak
        for i in range(len(experimental)):
            peak = <FittedPeak>PyList_GET_ITEM(experimental, i)
            peak.intensity *= factor

    cdef void scale_theoretical_peaks(self, list theoretical, double factor):
        cdef:
            size_t i
            TheoreticalPeak peak
        for i in range(len(theoretical)):
            peak = <TheoreticalPeak>PyList_GET_ITEM(theoretical, i)
            peak.intensity *= factor

    def evaluate(self, PeakSet peaklist, list observed, list expected):
        return self._evaluate(peaklist, observed, expected)

    cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):
        cdef:
            double score
        if self.scale_factor < 1:
            self.scale_factor = self._calculate_scale_factor(peaklist)
        self.scale_fitted_peaks(experimental, 1. / self.scale_factor)
        self.scale_theoretical_peaks(theoretical, 1. / self.scale_factor)
        score = self.scorer._evaluate(peaklist, experimental, theoretical)
        self.scale_fitted_peaks(experimental, self.scale_factor)
        self.scale_theoretical_peaks(theoretical, self.scale_factor)
        return score


cdef class DotProductFitter(IsotopicFitterBase):

    def __init__(self,  minimum_score=100):
        self.select = MaximizeFitSelector(minimum_score)

    cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):
        cdef:
            size_t i, n
            FittedPeak obs
            TheoreticalPeak theo
            double score
        n = PyList_GET_SIZE(experimental)
        score = 0
        for i in range(n):
            obs = <FittedPeak>PyList_GET_ITEM(experimental, i)
            theo = <TheoreticalPeak>PyList_GET_ITEM(theoretical, i)
            score += obs.intensity * theo.intensity
        return score



from .common import (
    ScanSerializerBase, ScanDeserializerBase, ProcessedRandomAccessScanSource)

from .mzml import (
    ProcessedMzMLDeserializer, ProcessedMzMLLoader, MzMLSerializer, ProcessedGeneric3DIonMobilityFrameSource, IonMobilityAware3DMzMLSerializer)

from .mzmlb import (
    MzMLbSerializer, ProcessedMzMLbDeserializer, ProcessedMzMLbLoader, IonMobilityAware3DMzMLbSerializer
)

from .text import (
    TextScanSerializerBase, HeaderedDelimitedWriter)

from .mgf import (
    ProcessedMGFDeserializer, MGFSerializer)


from .infer_type import (ProcessedMSFileLoader, get_writer)

__all__ = [
    "ScanSerializerBase",
    "ScanDeserializerBase",
    "ProcessedRandomAccessScanSource",
    "MzMLSerializer",
    "ProcessedMzMLDeserializer",
    "ProcessedMzMLLoader",
    "MGFSerializer",
    "ProcessedMGFDeserializer",
    "TextScanSerializerBase",
    "HeaderedDelimitedWriter",
    "ProcessedMSFileLoader",
    "get_writer",
    "IonMobilityAware3DMzMLSerializer",
    "ProcessedGeneric3DIonMobilityFrameSource",
]

if MzMLbSerializer is None:
    del MzMLbSerializer
else:
    __all__.extend([
        'MzMLbSerializer',
        'ProcessedMzMLbLoader',
        'ProcessedMzMLbDeserializer',
        'IonMobilityAware3DMzMLbSerializer',
    ])



from collections import namedtuple

import numpy as np

# from brainpy import neutral_mass as calc_neutral_mass
from ms_peak_picker import FittedPeak

from ms_deisotope.averagine import glycan, mass_charge_ratio
from ms_deisotope.scoring import g_test_scaled
from ms_deisotope.peak_set import DeconvolutedPeak, IonMobilityDeconvolutedPeak, IonMobilityProfileDeconvolutedPeakSolution, Envelope

from ms_deisotope.peak_dependency_network.intervals import SimpleInterval

from .shape_fitter import AdaptiveMultimodalChromatogramShapeFitter
from .lcms_feature import (
    EmptyFeature,
    LCMSFeature,
    LCMSFeatureTreeNode,
    RunningWeightedAverage,
    NodeFeatureSetIterator)


class map_coord(namedtuple("map_coord", ("mz", 'time'))):
    def __repr__(self):
        return "(%0.3f, %0.3f)" % self


class LCMSFeatureSetFit(object):
    def __init__(self, features, theoretical, score, charge,
                 missing_features=0, supporters=None, data=None,
                 neutral_mass=None, n_points=0, scores=None, times=None):
        if supporters is None:
            supporters = []
        if scores is None:
            scores = np.array([])
        if times is None:
            times = np.array([])
        self.features = features
        self.theoretical = theoretical
        self.score = score
        self.charge = charge
        self.data = data
        self.n_points = n_points
        self.missing_features = missing_features
        self.monoisotopic_feature = features[0]
        self.supporters = supporters
        self.mz = theoretical.monoisotopic_mz
        if neutral_mass is None:
            neutral_mass = neutral_mass(self.mz, self.charge)
        self.neutral_mass = neutral_mass
        self.scores = scores
        self.times = times

    def count_null_features(self):
        n_null = 0
        for feature in self.features:
            if feature is None or isinstance(feature, EmptyFeature):
                n_null += 1
        return n_null

    def has_multiple_real_features(self):
        return len(self) - self.count_null_features() > 1

    def clone(self):
        return self.__class__(
            self.features, self.theoretical, self.score, self.charge,
            self.missing_features, self.supporters, self.data,
            self.neutral_mass, self.scores, self.times)

    def __reduce__(self):
        return self.__class__, (
            self.features, self.theoretical, self.score, self.charge,
            self.missing_features, self.supporters, self.data, self.neutral_mass,
            self.scores, self.times)

    def __eq__(self, other):
        val = (self.score == other.score and
               self.charge == other.charge and
               self.features == other.features and
               self.theoretical == other.theoretical)
        if self.data is not None or other.data is not None:
            val = val and (self.data == other.data)
        return val

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        return self.score < other.score

    def __gt__(self, other):
        return self.score > other.score

    def __hash__(self):
        return hash((self.monoisotopic_feature.mz, self.charge))

    def __iter__(self):
        return iter(self.features)

    def __len__(self):
        return len(self.features)

    @property
    def npeaks(self):
        return len(self)

    def __repr__(self):
        return "LCMSFeatureSetFit(score=%0.5f, charge=%d, size=%d, monoisotopic_mz=%0.5f, %0.2f-%0.2f)" % (
            self.score, self.charge, len(self), self.monoisotopic_feature.mz,
            self.start.time, self.end.time)

    @property
    def start(self):
        first = self.features[0]
        if first is None:
            raise Exception()
        return map_coord(first.mz, first.start_time)

    @property
    def end(self):
        for last in reversed(self.features):
            if last is None:
                continue
            return map_coord(last.mz, last.end_time)


class DeconvolutedLCMSFeatureTreeNode(LCMSFeatureTreeNode):
    __slots__ = ["_neutral_mass", "charge", "precursor_information"]

    def __init__(self, time=None, members=None, precursor_information=None):
        if precursor_information is None:
            precursor_information = []
        self._neutral_mass = 0
        self.charge = 0
        super(DeconvolutedLCMSFeatureTreeNode, self).__init__(time, members)
        self.precursor_information = precursor_information

    def _recalculate(self):
        self._calculate_most_abundant_member()
        if self._most_abundant_member is not None:
            self._mz = self._most_abundant_member.mz
            self._neutral_mass = self._most_abundant_member.neutral_mass
            self.charge = self._most_abundant_member.charge

    @property
    def neutral_mass(self) -> float:
        if self._neutral_mass == 0:
            if self._most_abundant_member is not None:
                self._neutral_mass = self._most_abundant_member.neutral_mass
        return self._neutral_mass


class DeconvolutedLCMSFeature(LCMSFeature):
    __slots__ = ('score', 'n_features', 'supporters',
                 '_neutral_mass', '_last_neutral_mass',
                 '_precursor_information', 'charge', '_mz')

    def __init__(self, nodes=None, charge=None, adducts=None, used_as_adduct=None, score=0.0,
                 n_features=0, feature_id=None, supporters=None):
        if supporters is None:
            supporters = []
        self.charge = charge
        self.score = score
        self._neutral_mass = None
        self._last_neutral_mass = None
        self._precursor_information = None
        self._mz = None
        self.n_features = n_features
        self.supporters = supporters
        super(DeconvolutedLCMSFeature, self).__init__(nodes, adducts, used_as_adduct, feature_id=feature_id)

    def _initialize_averager(self):
        self._peak_averager = DeconvolutedRunningWeightedAverage()

    def __reduce__(self):
        return self.__class__, (self.nodes, self.charge, self.adducts,
                                self.used_as_adduct, self.score, self.feature_id, self.supporters)

    @property
    def precursor_information(self):
        if self._precursor_information is None:
            pinfo = []
            for node in self:
                pinfo.extend(node.precursor_information)
            self._precursor_information = tuple(pinfo)
        return self._precursor_information

    def clone(self, deep=False, cls=None) -> 'DeconvolutedLCMSFeature':
        if cls is None:
            cls = self.__class__
        return cls(
            self.nodes.clone(deep=deep), self.charge, self.adducts, self.used_as_adduct, self.score,
            self.n_features, self.feature_id, list(self.supporters))

    def _invalidate(self, reaverage: bool=False):
        self._last_neutral_mass = self._neutral_mass if self._neutral_mass is not None else 0.
        self._neutral_mass = None
        self._precursor_information = None
        self._mz = None
        super(DeconvolutedLCMSFeature, self)._invalidate(reaverage)

    def _update_from_averager(self):
        self._neutral_mass = self._last_neutral_mass = self._peak_averager.current_mean

    @property
    def neutral_mass(self) -> float:
        if self._neutral_mass is None:
            self._reaverage_and_update()
        return self._neutral_mass

    @property
    def mz(self):
        if self._mz is None:
            self._mz = mass_charge_ratio(self.neutral_mass, self.charge)
        return self._mz

    def _copy_chunk(self, nodes, *args, **kwargs) -> 'DeconvolutedLCMSFeature':
        x = self.__class__(
            nodes, self.charge, list(self.adducts), list(self.used_as_adduct),
            self.score, self.n_features, None, list(self.supporters))
        return x

    def sum(self, other):
        missing = []
        feat_iter = NodeFeatureSetIterator([self, other])
        for nodes in feat_iter:
            base = nodes[0]
            new = nodes[1]
            if base is None:
                missing.append(new)
            elif new is not None:
                base.members[0].intensity += new.members[0].intensity
                base.precursor_information.extend(new.precursor_information)
        if missing:
            for node in missing:
                self.insert_node(DeconvolutedLCMSFeatureTreeNode(
                    node.time, list(node.members), list(node.precursor_information)))
        self.supporters.extend(other.supporters)
        return self

    def __repr__(self):
        return "%s(%0.4f, %d, %0.2f, %0.2f, %0.2f)" % (
            self.__class__.__name__, self.neutral_mass,
            self.charge, self.score,
            self.start_time, self.end_time)

    def sum_envelopes(self) -> Envelope:
        mzs = None
        intensities = None
        for node in self:
            peak: DeconvolutedPeak
            for peak in node.members:
                if mzs is None:
                    mzs = [0.0 for i in range(len(peak.envelope))]
                    intensities = [0.0 for i in range(len(peak.envelope))]
                for i, pt in enumerate(peak.envelope):
                    try:
                        mzs[i] += (pt.mz * pt.intensity)
                        intensities[i] += pt.intensity
                    except IndexError:
                        mzs.append(pt.mz * pt.intensity)
                        intensities.append(pt.intensity)
        out = [None for i in range(len(mzs))]
        for i, mz_int in enumerate(mzs):
            inten = intensities[i]
            out[i] = (mz_int / inten, inten)
        return Envelope(out)


class DeconvolutedRunningWeightedAverage(RunningWeightedAverage):
    def add(self, peak: DeconvolutedPeak):
        if peak.intensity == 0:
            if self.current_mean == 0 and self.total_weight == 0:
                self.current_mean = peak.neutral_mass
                self.total_weight = 1
            else:
                return
        agg = (self.total_weight * self.current_mean) + \
            (peak.neutral_mass * peak.intensity)
        self.total_weight += peak.intensity
        self.current_mean = agg / self.total_weight
        self.current_count += 1
        return self

try:
    from ms_deisotope._c.feature_map.lcms_feature import RunningWeightedAverageNeutralMass as DeconvolutedRunningWeightedAverage
except ImportError:
    pass


class DeconvolutedDriftTimeRunningWeightedAverage(DeconvolutedRunningWeightedAverage):
    def _initialize(self):
        self.neutral_mass_mean = 0.0
        self.drift_time_mean = 0.0

    def add(self, peak: IonMobilityDeconvolutedPeak):
        if peak.intensity == 0:
            if self.current_mean == 0 and self.total_weight == 0:
                self.neutral_mass_mean = peak.neutral_mass
                self.drift_time_mean = peak.drift_time
                self.total_weight = 1
            else:
                return
        drift_time_agg = (self.total_weight * self.drift_time_mean) + \
            (peak.drift_time * peak.intensity)
        neutral_mass_agg = (self.total_weight * self.neutral_mass_mean) + \
            (peak.neutral_mass * peak.intensity)
        self.total_weight += peak.intensity
        self.drift_time_mean = drift_time_agg / self.total_weight
        self.neutral_mass_mean = neutral_mass_agg / self.total_weight
        return self


class IonMobilityDeconvolutedLCMSFeature(DeconvolutedLCMSFeature):
    def __init__(self, nodes=None, charge=None, adducts=None, used_as_adduct=None, score=0.0,
                 n_features=0, feature_id=None, supporters=None):
        self._drift_time = None
        self._last_drift_time = None
        super(IonMobilityDeconvolutedLCMSFeature, self).__init__(
            nodes=nodes, charge=charge, adducts=adducts, used_as_adduct=used_as_adduct, score=score,
            n_features=n_features, feature_id=feature_id, supporters=supporters)

    def _initialize_averager(self):
        self._peak_averager = DeconvolutedDriftTimeRunningWeightedAverage()

    def _invalidate(self, reaverage: bool=False):
        self._last_drift_time = self._drift_time if self._drift_time is not None else 0.
        self._drift_time = None
        return super(IonMobilityDeconvolutedLCMSFeature, self)._invalidate(reaverage=reaverage)

    def _update_from_averager(self):
        self._neutral_mass = self._last_neutral_mass = self._peak_averager.neutral_mass_mean
        self._drift_time = self._last_drift_time = self._peak_averager.drift_time_mean

    @property
    def drift_time(self) -> float:
        if self._drift_time is None:
            self._reaverage_and_update()
        return self._drift_time

    def __repr__(self):
        return "%s(%0.4f, %0.4f, %d, %0.2f, %0.2f, %0.2f)" % (
            self.__class__.__name__, self.neutral_mass, self.drift_time,
            self.charge, self.score,
            self.start_time, self.end_time)


def envelope_to_peak_list(envelope):
    return [FittedPeak(e[0], e[1], 0, 0, 0, 0, 0, 0, 0) for e in envelope]


def scale_theoretical_isotopic_pattern(eid, tid):
    total = sum(p.intensity for p in eid)
    for p in tid:
        p.intensity *= total


def isotopic_consistency(eic, averagine=glycan, truncate_after=0.95):
    peak_scores = []
    peak_abundances = []
    for node in eic:
        for peak in node.members:
            eid = envelope_to_peak_list(peak.envelope)
            tid = averagine.isotopic_cluster(peak.mz, peak.charge, truncate_after=truncate_after)
            tid.scale(eid)
            peak_scores.append(abs(g_test_scaled(None, eid, tid.truncated_tid)))
            peak_abundances.append(peak.intensity)
    return max(1 - np.average(peak_scores, weights=peak_abundances), 1e-4)


def spacing_fit(eic):
    times, intensities = eic.as_arrays()
    last_rt = times[0]
    last_int = intensities[0]
    rt_deltas = []
    intensity_deltas = []
    for rt, inten in zip(times[1:], intensities[1:]):
        d_rt = rt - last_rt
        rt_deltas.append(d_rt)
        intensity_deltas.append(abs(last_int - inten))
        last_rt = rt
        last_int = inten
    return max(1 - np.average(rt_deltas, weights=intensity_deltas) * 2, 1e-4)


def shape_fit(eic, smooth=0.15):
    return max(1 - AdaptiveMultimodalChromatogramShapeFitter(eic, smooth=smooth).line_test, 1e-4)


def profile_qc(eic, smooth=0.15, averagine=glycan, truncate_after=0.95):
    v = 1.0
    v *= isotopic_consistency(eic, averagine, truncate_after)
    v *= spacing_fit(eic)
    v *= shape_fit(eic, smooth)
    return v


try:
    has_c = True
    _map_coord = map_coord
    _LCMSFeatureSetFit = LCMSFeatureSetFit
    from ms_deisotope._c.feature_map.feature_fit import (LCMSFeatureSetFit, map_coord)
except ImportError as e:
    print(e)
    has_c = False


class IonMobilityProfileDeconvolutedRunningWeightedAverage(DeconvolutedRunningWeightedAverage):
    def __init__(self, *args, **kwargs):
        self._initialize()

    def _initialize(self):
        super()._initialize()
        self.interval = SimpleInterval(0, 0)

    def add(self, peak: IonMobilityProfileDeconvolutedPeakSolution):
        if peak.intensity == 0:
            if self.current_mean == 0 and self.total_weight == 0:
                self.current_mean = peak.neutral_mass
                self.interval.start = peak.ion_mobility_interval.start
                self.interval.end = peak.ion_mobility_interval.end
                self.total_weight = 1
            else:
                return
        agg = (self.total_weight * self.current_mean) + \
            (peak.neutral_mass * peak.intensity)
        start_agg = (self.interval.start * self.total_weight) + (peak.ion_mobility_interval.start * peak.intensity)
        end_agg = (self.interval.end * self.total_weight) + (peak.ion_mobility_interval.end * peak.intensity)
        self.total_weight += peak.intensity
        self.current_mean = agg / self.total_weight
        self.interval.start = start_agg / self.total_weight
        self.interval.end = end_agg / self.total_weight
        self.current_count += 1
        return self


class IonMobilityProfileDeconvolutedLCMSFeature(DeconvolutedLCMSFeature):
    __slots__ = ('_ion_mobility_interval', '_last_ion_mobility_interval')

    _ion_mobility_interval: SimpleInterval

    def __init__(self, nodes=None, charge=None, adducts=None, used_as_adduct=None, score=0.0,
                 n_features=0, feature_id=None, supporters=None):
        super().__init__(nodes, charge, adducts, used_as_adduct, score,
                         n_features, feature_id, supporters)
        self._ion_mobility_interval = None
        self._last_ion_mobility_interval = None

    def _invalidate(self, reaverage: bool=False):
        self._last_ion_mobility_interval = self._ion_mobility_interval if self._ion_mobility_interval is not None \
            else SimpleInterval(0, 0)
        self._ion_mobility_interval = None
        super()._invalidate(reaverage)

    def _initialize_averager(self):
        self._peak_averager = IonMobilityProfileDeconvolutedRunningWeightedAverage()

    def _update_from_averager(self):
        self._neutral_mass = self._last_neutral_mass = self._peak_averager.current_mean
        self._ion_mobility_interval = self._last_ion_mobility_interval = self._peak_averager.interval

    @property
    def neutral_mass(self) -> float:
        if self._neutral_mass is None:
            self._reaverage_and_update()
        return self._neutral_mass

    @property
    def ion_mobility_interval(self) -> SimpleInterval:
        if self._ion_mobility_interval is None:
            self._reaverage_and_update()
        return self._ion_mobility_interval

    def between_ion_mobilities(self, ion_mobility_start: float, ion_mobility_end: float,
                               time_start: float = 0, time_end: float = float('inf'),
                               include_mz: bool=True, include_envelope: bool=True) -> 'IonMobilityProfileDeconvolutedLCMSFeature':
        nodes = []
        node, i = self.find_time(time_start)
        if i > 0:
            i -= 1
        for node in self:
            if node.time < time_start:
                continue
            if node.time > time_end:
                break
            peaks = []
            for peak in node.members:
                try:
                    peaks.append(
                        peak.from_feature(
                            peak.solution.between_times(ion_mobility_start, ion_mobility_end),
                            include_mz=include_mz, include_envelope=include_envelope))
                except IndexError:
                    continue
            if peaks:
                node = node.__class__(node.time, peaks)
                nodes.append(node)
        return self._copy_chunk(nodes)


try:
    has_c = True
    from ms_deisotope._c.feature_map.feature_fit import (
        _sum_envelopes as _c_sum_envelopes)

    DeconvolutedLCMSFeature.sum_envelopes = _c_sum_envelopes
except ImportError as e:
    print(e)
    has_c = False



from ms_deisotope._c.scoring cimport IsotopicFitRecord
from ms_peak_picker._c.peak_set cimport PeakBase


cdef class _Index:
    cdef:
        public size_t neutral_mass
        public size_t mz

    @staticmethod
    cdef _Index _create(size_t neutral_mass, size_t mz)


cdef class EnvelopePair:
    cdef:
        public double mz
        public double intensity

    cpdef bint _eq(self, EnvelopePair other)

    @staticmethod
    cdef EnvelopePair _create(double mz, double intensity)


cdef class Envelope:
    cdef public tuple pairs

    cdef inline size_t get_size(self)

    cdef inline EnvelopePair getitem(self, size_t i)

    cpdef bint _eq(self, Envelope other)

    cpdef Envelope clone(self)

    @staticmethod
    cdef Envelope _create(tuple pairs)


cdef class DeconvolutedPeak(PeakBase):
    cdef:
        public double neutral_mass
        # public double intensity
        public double signal_to_noise
        public _Index _index
        public double full_width_at_half_max
        public int charge
        public double a_to_a2_ratio
        public double most_abundant_mass
        public double average_mass
        public double score
        # public double area
        public Envelope envelope
        # public double mz
        public IsotopicFitRecord fit
        public bint chosen_for_msms

    cpdef bint _eq(self, DeconvolutedPeak other)

    @staticmethod
    cdef DeconvolutedPeak _create_simple(
        double neutral_mass, double intensity, int charge, double score,
        double mz, Envelope envelope)


cdef class IonMobilityDeconvolutedPeak(DeconvolutedPeak):
    cdef:
        public double drift_time


cdef class DeconvolutedPeakSolution(DeconvolutedPeak):
    cdef:
        public object solution

cdef class DeconvolutedPeakSet(object):
    cdef:
        public tuple peaks
        public tuple _mz_ordered
        public bint indexed


    cpdef reindex(self)

    cpdef DeconvolutedPeakSet clone(self)
    cpdef DeconvolutedPeakSet copy(self)

    cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)

    cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=*, bint use_mz=*)

    cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=*)

    cdef DeconvolutedPeak _get_nearest_peak(self, double neutral_mass, double* errout)

    cdef size_t get_size(self)

    cdef DeconvolutedPeak getitem(self, size_t i)
    cdef tuple getslice(self, size_t start, size_t end)


cdef int _binary_search_interval(double* array, double target, double error_tolerance, size_t n, size_t* start, size_t* end) nogil


cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):
    cdef:
        double* neutral_mass_array
        double* mz_array
        index_list* interval_index

        size_t _size

    cdef void _build_index_arrays(self)
    cdef int _interval_for(self, double neutral_mass, double tolerance, size_t* start, size_t* end) nogil


cdef size_t INTERVAL_INDEX_SIZE

cdef struct index_cell:
    double center_value
    size_t start
    size_t end


cdef struct index_list:
    index_cell* index
    size_t size
    double low
    double high


cdef struct deconvoluted_peak_t:
    double neutral_mass
    double intensity
    int charge
    unsigned int index


cdef enum CPeaksFlags:
    borrowing = 0
    owns_peaks = 1
    owns_index = 2
    owns_peaks_and_index = 3

cdef struct deconvoluted_peak_set_t:
    deconvoluted_peak_t* peaks
    double* mass_index
    size_t size
    CPeaksFlags flags


cdef int create_deconvoluted_peak_set_t(DeconvolutedPeakSetIndexed peak_set, deconvoluted_peak_set_t* destination) except 1
cdef int free_deconvoluted_peak_set_t(deconvoluted_peak_set_t* destination) nogil

cdef deconvoluted_peak_set_t deconvoluted_peak_set_all_peaks_for(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil
cdef deconvoluted_peak_t* deconvoluted_peak_set_has_peak(deconvoluted_peak_set_t* self, double neutral_mass, double error_tolerance=*) nogil

cdef int deconvoluted_peak_eq(deconvoluted_peak_t* self, deconvoluted_peak_t* other) nogil
cdef size_t deconvoluted_peak_hash(deconvoluted_peak_t* self) nogil


cdef class _CPeakSet:
    cdef:
        deconvoluted_peak_set_t* ptr

    cdef deconvoluted_peak_t* getitem(self, size_t i) nogil

    cdef deconvoluted_peak_t* _has_peak(self, double neutral_mass, double error_tolerance=*) nogil
    cdef deconvoluted_peak_set_t _all_peaks_for(self, double neutral_mass, double error_tolerance=*) nogil

    cpdef object has_peak(self, double neutral_mass, double error_tolerance=*)
    cpdef tuple all_peaks_for(self, double neutral_mass, double error_tolerance=*)



from .common import ScanSerializerBase


class _HeaderInformation(dict):
    """A simple key-value store that also carries a reference to a :class:`~.Scan` object.

    Intended to be used internally by the :class:`TextScanSerializerBase` class hierarchy,
    but should not be needed elsewhere either.

    """
    def __init__(self, scan):
        dict.__init__(self)
        self.scan = scan


class TextScanSerializerBase(ScanSerializerBase):
    """A base class for text file-based scan serialization formats like
    simple custom tabular formats, MS1, MS2, or MGF.

    This class tries to flatten out the hierarchical structure of a :class:`~.Scan`
    into a dictionary that is more convenient to project onto a text file.

    Attributes
    ----------
    stream: writable file-like
        The stream to write content to. Should support basic IO operations like
        write, flush, and close.
    deconvoluted: bool
        Whether the file is storing spectra in deconvoluted mode or not.

    """
    def __init__(self, stream, deconvoluted=True):
        super(TextScanSerializerBase, self).__init__(stream, deconvoluted=deconvoluted)
        self.stream = stream
        self.deconvoluted = deconvoluted

    def close(self):
        """Close :attr:`stream`.
        """
        self.stream.close()

    def save_scan(self, scan, **kwargs):
        self.write_scan(*self.prepare_scan_data(scan))

    def construct_header(self, scan):
        """Build a dictionary of properties describing `scan` specifically not
        including the peak list or signal information.

        Parameters
        ----------
        scan : :class:`~.ScanBase`
            The scan to describe

        Returns
        -------
        :class:`_HeaderInformation`
        """
        prec_info = scan.precursor_information
        header_dict = _HeaderInformation(scan)
        if prec_info is not None:
            if prec_info.extracted_charge:
                header_dict['precursor_neutral_mass'] = prec_info.extracted_neutral_mass
                header_dict['precursor_mz'] = prec_info.extracted_mz
                header_dict['precursor_charge'] = prec_info.extracted_charge
                header_dict['precursor_intensity'] = prec_info.extracted_intensity
            else:
                header_dict['precursor_neutral_mass'] = prec_info.neutral_mass
                header_dict['precursor_mz'] = prec_info.mz
                header_dict['precursor_charge'] = prec_info.charge
                header_dict['precursor_intensity'] = prec_info.intensity
            header_dict['precursor_scan_id'] = prec_info.precursor_scan_id
            header_dict['defaulted'] = prec_info.defaulted
            activation = scan.activation
            if activation:
                header_dict['precursor_activation_method'] = activation.method
                header_dict['precursor_activation_energy'] = activation.energy
                if activation.is_multiple_dissociation():
                    header_dict['precursor_activation_method'] = ';'.join([str(m) for m in activation.methods])
                    header_dict['precursor_activation_energy'] = ';'.join([str(m) for m in activation.energies])
                for key, value in activation.data.items():
                    header_dict['precursor_activation_' + key.lower()] = value
        header_dict['scan_time'] = scan.scan_time
        header_dict['id'] = scan.id
        header_dict['title'] = scan.title
        header_dict['ms_level'] = scan.ms_level
        header_dict['polarity'] = scan.polarity
        header_dict['index'] = scan.index
        instrument_config = scan.instrument_configuration
        if instrument_config:
            header_dict['analyzers'] = ';'.join(a.name for a in instrument_config.analyzers)
        if scan.has_ion_mobility():
            header_dict['drift_time'] = scan.drift_time
        header_dict['annotations'] = scan.annotations
        return header_dict

    def format_peak_vectors(self, scan):
        """Build a simple peak attribute list format commonly used by most
        text formats.

        Most text based formats represent peaks as a list of delimited values
        per line. This method unpacks peak lists into parallel lists for each
        peak attribute.

        This method may behave differently if :attr:`deconvoluted` is true or
        not.

        Parameters
        ----------
        scan : :class:`~.ScanBase`
            The scan to extract peak attributes for.

        Returns
        -------
        mz: list of float
            The m/z of each peak. If :attr:`deconvoluted` is true, this may be
            the peak's neutral mass.
        intensity: list of float
            The intensity of each peak
        charge: list of int, optional
            If :attr:`deconvoluted` is true, this may be a list of each peak's charge
            state. :const:`None` otherwise.
        """
        if self.deconvoluted:
            neutral_mass = [p.neutral_mass for p in scan.deconvoluted_peak_set]
            intensity = [p.intensity for p in scan.deconvoluted_peak_set]
            charge = [p.charge for p in scan.deconvoluted_peak_set]
            return (neutral_mass, intensity, charge)
        else:
            mz = [p.mz for p in scan.peak_set]
            intensity = [p.intensity for p in scan.peak_set]
            return mz, intensity, None

    def prepare_scan_data(self, scan):
        """Prepare data from `scan` for formatting as text.

        Parameters
        ----------
        scan: :class:`~.ScanBase`
            The scan to prepare

        Returns
        -------
        header: :class:`_HeaderInformation`
            The scan header metadata
        vectors: :class:`tuple` of :class:`list`
            The peak attribute lists for all peaks in `scan`
        """
        header_dict = self.construct_header(scan)
        vectors = self.format_peak_vectors(scan)
        return (header_dict, vectors)

    def write_scan(self, scan_header, data_vectors):
        """Format the given scan data and write it to :attr:`stream`

        Should receive its parameter information from :meth:`prepare_scan_data`

        Parameters
        ----------
        scan_header : :class:`_HeaderInformation`
            The scan header metadata
        data_vectors : :class:`Sequence`
            The peak attribute lists

        """
        raise NotImplementedError()


class HeaderedDelimitedWriter(TextScanSerializerBase):
    """A simple but usable implementation of :class:`TextScanSerializerBase`

    Writes all metadata headers as "#{key}={value}\\n" immediately preceding
    the peak attribute list.

    """
    def __init__(self, stream, deconvoluted=True):
        super(HeaderedDelimitedWriter, self).__init__(stream, deconvoluted)

    def write_header(self, header_dict):
        """Write the header information in the form "#{key}={value}" with one
        key-value pair per line.

        Parameters
        ----------
        header_dict : :class:`_HeaderInformation`
        """
        for key, value in header_dict.items():
            self.stream.write(("#%s=%s\n" % (key, value)).encode("utf8"))

    def write_vectors(self, vectors):
        """Write out the peak attribute lists along peak index.

        Parameters
        ----------
        vectors : :class:`Sequence`
            The peak attribute lists to write.
        """
        vectors = [v for v in vectors if v is not None]
        if sum(map(len, vectors)) != len(vectors[0]) * len(vectors):
            raise ValueError("All data vectors must be equal!")
        for i in range(len(vectors[0])):
            line = []
            for j, vector in enumerate(vectors):
                if j != 0:
                    line.append(' ')
                line.append(str(vector[i]))
            line.append("\n")
            self.stream.write(''.join(line).encode('utf8'))

    def write_scan(self, scan_header, data_vectors):
        self.write_header(scan_header)
        self.write_vectors(data_vectors)



from pyteomics.usi import USI, proxi, _PROXIBackend

from .scan import ScanDataSource, Scan, RawDataArrays, PrecursorInformation, ChargeNotProvided
from .metadata.scan_traits import IsolationWindow


USI_KEY = "_$USI"

class _PROXIScanSource(ScanDataSource):

    def _scan_arrays(self, scan):
        return RawDataArrays(scan.get("m/z array"), scan.get('intensity array'))

    def _scan_id(self, scan):
        if USI_KEY in scan:
            return scan[USI_KEY]
        if 'usi' in scan:
            return scan['usi']
        attributes = scan['attributes']
        for attrib in attributes:
            if attrib['name'] == 'scan number':
                return attrib['value']

    def _scan_index(self, scan):
        attributes = scan['attributes']
        for attrib in attributes:
            if attrib['name'] == 'scan number':
                return int(attrib['value']) - 1
        return -1

    def _scan_title(self, scan):
        if USI_KEY in scan:
            return scan[USI_KEY]
        if 'usi' in scan:
            return scan['usi']
        attributes = scan['attributes']
        for attrib in attributes:
            if attrib['name'] == 'scan number':
                return attrib['value']

    def _scan_time(self, scan):
        return None

    def _ms_level(self, scan):
        return 2

    def _precursor_information(self, scan):
        mz = 0.0
        charge = ChargeNotProvided
        attributes = scan['attributes']
        for attrib in attributes:
            if attrib['name'] == 'isolation window target m/z':
                mz = float(attrib['value'])
        return PrecursorInformation(mz, 0, charge, source=self)

    def _is_profile(self, scan):
        # TODO: Not guaranteed but no counter-examples
        return False

    def _polarity(self, scan):
        # TODO: Not guaranteed but no counter-examples
        return 1

    def _activation(self, scan):
        return None


class PROXIServiceClient(_PROXIScanSource):
    """A :class:`~.ScanDataSource` client that provides random access to individual spectra
    by USI.

    Attributes
    ----------
    backend : str or :class:`pyteomics.usi._PROXIBackend`
        The server to fetch spectra from. If a name is given, the server will
        be inferred, else if a URL template is given, a new :class:`pyteomics.usi._PROXIBackend`
        is created from it.
    """
    def __init__(self, backend='peptide_atlas'):
        self.backend = self._make_backend(backend)

    def _make_backend(self, backend):
        if isinstance(backend, _PROXIBackend):
            return backend
        if isinstance(backend, str):
            # We're given a URL
            if backend.startswith("http"):
                return _PROXIBackend(backend, backend)
        return backend

    def get_scan_by_id(self, scan_id):
        """Get a spectrum by USI from the server.

        Parameters
        ----------
        scan_id : str
            The USI to fetch.


        Returns
        -------
        Scan
        """
        payload = proxi(scan_id, self.backend)
        payload[USI_KEY] = scan_id
        return self._make_scan(payload)

    def get(self, usi):
        """Get a spectrum by USI from the server.

        Parameters
        ----------
        scan_id : str
            The USI to fetch.


        Returns
        -------
        Scan
        """
        return self.get_scan_by_id(usi)

    def __call__(self, usi):
        """Get a spectrum by USI from the server.

        Parameters
        ----------
        scan_id : str
            The USI to fetch.


        Returns
        -------
        Scan
        """
        return self.get_scan_by_id(usi)

    def __repr__(self):
        return "{self.__class__.__name__}({self.backend})".format(self=self)


peptide_atlas = PROXIServiceClient("peptide_atlas")
massive = PROXIServiceClient("massive")
pride = PROXIServiceClient("pride")

parse = USI.parse
